#include "dll_dependency.h"

static OrderedUniqueList *SystemDllDirectories;
static OrderedUniqueList *DllDirectories;

static void toLower(String *string) {
    for (char *str = (char *) string->cStr; *str != '\0'; str++)
        if (*str >= 'A' && *str <= 'Z')
            *str |= 0x20;
}

static bool tryFindDllPath(DllDependency *dependency) {
    OrderedUniqueList *systemDllDirectories = useSystemDllDirectories();
    OrderedUniqueList *dllDirectories = useDllDirectories();

    for (int index = 0; index < systemDllDirectories->length; index++) {
        String *sysDir = (String *) systemDllDirectories->data[index];
        defer String *pathCandidate = string("%s/%s", sysDir->cStr, dependency->dllName->cStr);

        if (access(pathCandidate->cStr, F_OK) == 0) {
            dependency->isSystem = true;
            dependency->dllPath = stringDup(pathCandidate);
            return true;
        }
    }

    for (int index = 0; index < dllDirectories->length; index++) {
        String *dllDir = (String *) dllDirectories->data[index];
        defer String *pathCandidate = string("%s/%s", dllDir->cStr, dependency->dllName->cStr);

        if (access(pathCandidate->cStr, F_OK) == 0) {
            dependency->dllPath = stringDup(pathCandidate);
            return true;
        }
    }

    return false;
}

DllDependency *newDllDependency(const String *dllName) {
    DllDependency *dependency = dalloc(DllDependency);
    dependency->dllName = stringDup(dllName);
    dependency->symbols = newOrderedUniqueList((ComparatorFunc) compareString);

    if (!tryFindDllPath(dependency)) {
        toLower(dependency->dllName);
        tryFindDllPath(dependency);
    }

    checkError(dependency, dependency->dllPath != NULL, "DLL does not exist: %s", dependency->dllName->cStr);
    return dependency;
}

int compareDllDependency(DllDependency *a, DllDependency *b) {
    return strncmp(a->dllName->cStr, b->dllName->cStr, min(a->dllName->length, b->dllName->length));
}

void freeDllDependency(DllDependency *dependency) {
    if (dependency->dllName != NULL) freeDeferable(dependency->dllName);
    if (dependency->dllPath != NULL) freeDeferable(dependency->dllPath);
    if (dependency->symbols != NULL) freeDeferable(dependency->symbols);
}

OrderedUniqueList *useSystemDllDirectories() {
    if (SystemDllDirectories == NULL) {
        const char *winePrefix = getenv("WINEPREFIX");
        SystemDllDirectories = newOrderedUniqueList((ComparatorFunc) compareString);

        if (winePrefix != NULL) {
            listInsert(SystemDllDirectories, string("%s/drive_c/windows/system32", winePrefix));
        } else {
            listInsert(SystemDllDirectories, string("%s/.wine/drive_c/windows/system32", getenv("HOME")));
        }
    }

    return SystemDllDirectories;
}

OrderedUniqueList *useDllDirectories() {
    if (DllDirectories == NULL) {
        DllDirectories = newOrderedUniqueList((ComparatorFunc) compareString);
        listInsert(DllDirectories, string("/usr/x86_64-w64-mingw32/bin"));

        if (getenv("WINEDLLPATH") != NULL) {
            defer char *wineDllPath = strdup(getenv("WINEDLLPATH"));
            char *token = strtok(wineDllPath, ":");

            while (token != NULL) {
                listInsert(DllDirectories, string("%s", token));
                token = strtok(NULL, ":");
            }
        }
    }

    return DllDirectories;
}
