#define VH_IMPLEMENTATION

#include <string.h>
#include <libgen.h>
#include <limits.h>
#include <stdlib.h>
#include "arguments.h"
#include "mapped_file.h"
#include "pe_header.h"
#include "section.h"
#include "ordered_unique_list.h"
#include "dll_dependency.h"
#include "vh.h"
#include "import_lookup_table.h"

#define UnknownLoc (useArguments()->colorOutput ? unknownLocationColored : unknownLocation)
const char *unknownLocation = "(unknown location)";
const char *unknownLocationColored = "\033[1;33m(unknown location)\033[0m";

typedef struct {
    uint32_t iltOffset;
    uint32_t dateTimeStamp;
    uint32_t forwarderChain;
    uint32_t nameOffset;
    uint32_t iatOffset;
} ImportTable;

typedef struct {
    uint32_t padding0;
    uint32_t nameOffset;
    uint32_t moduleHandle;
    uint32_t delayIat;
    uint32_t delayNameOffset;
    uint32_t boundDelayTableOffset;
    uint32_t unloadDelayTableOffset;
    uint32_t timeStamp;
} DelayedImportTable;

static void resolveDependencies(OrderedUniqueList *dependencyList, const String *fileName) {
    defer MappedFile *peFile = newMappedFile(fileName->cStr);
    fatal(peFile, "Unable to open PE file at " SFMT ": %s\n", SALC(fileName), errStr(peFile));

    defer PeHeader *peHeader = parsePeHeader(peFile);
    fatal(peHeader, "Unable to parse PE header or " SFMT ": %s", SALC(fileName), errStr(peHeader));
    const DataDirectories *dataDir = getDataDirectories(peHeader);

    defer Section *importsSection = findSectionByRva(
            peHeader, (long) dataDir->importTable.offset
    );

    checkElse(importsSection, "PE file at " SFMT " has no imports: %s", SALC(fileName), errStr(importsSection)) {
        const ImportTable *importTable = resolveSectionRva(
                importsSection, (long) dataDir->importTable.offset
        );

        for (; importTable->nameOffset != 0; importTable++) {
            defer String *dllName = string(resolveSectionRva(importsSection, importTable->nameOffset));
            DllDependency *dependency = newDllDependency(dllName);

            if (useArguments()->errUnknownLocation)
                fatal(dependency, "Unable to resolve dependency %s: %s", dllName->cStr, errStr(dependency));

            bool isNewDependency = listInsertOrFetch(dependencyList, &dependency);
            check(dependency) dependency->dllPath = string(UnknownLoc);
            else if (isNewDependency && !dependency->isSystem && useArguments()->recursive)
                resolveDependencies(dependencyList, dependency->dllPath);

            if (useArguments()->json) {
                defer ImportLookupTable *ilt = newImportLookupTable(peHeader, importTable->iltOffset);

                while (hasNextHint(ilt)) {
                    defer HintNameTable *hint = getNextHint(ilt);
                    listInsert(dependency->symbols, string(hint->name));
                }
            }
        }
    }

    defer Section *delayedImportsSection = findSectionByRva(
            peHeader, (long) dataDir->delayImportDescriptor.offset
    );

    noError(delayedImportsSection) {
        const DelayedImportTable *delayedImportTable = resolveSectionRva(
                delayedImportsSection, (long) dataDir->delayImportDescriptor.offset
        );

        for (; delayedImportTable->nameOffset != 0; delayedImportTable++) {
            defer String *dllName = string(resolveSectionRva(importsSection, delayedImportTable->nameOffset));
            DllDependency *dependency = newDllDependency(dllName);
            dependency->isDelayed = true;

            if (useArguments()->errUnknownLocation)
                fatal(dependency, "Unable to resolve dependency %s: %s", dllName->cStr, errStr(dependency));

            bool isNewDependency = listInsertOrFetch(dependencyList, &dependency);
            check(dependency) dependency->dllPath = string(UnknownLoc);
            else if (isNewDependency && !dependency->isSystem && useArguments()->recursive)
                resolveDependencies(dependencyList, dependency->dllPath);

            if (useArguments()->json) {
                defer ImportLookupTable *ilt = newImportLookupTable(peHeader, delayedImportTable->delayNameOffset);

                while (hasNextHint(ilt)) {
                    defer HintNameTable *hint = getNextHint(ilt);
                    listInsert(dependency->symbols, string(hint->name));
                }
            }
        }
    }
}

int main(int argc, char **argv) {
    Arguments *arguments = parseArguments(argc, argv);
    defer String *pePath = string(arguments->peFilePath);
    defer OrderedUniqueList *list = newOrderedUniqueList((ComparatorFunc) compareDllDependency);

    char dirPath[PATH_MAX] = {0};
    realpath(arguments->peFilePath, dirPath);
    dirname(dirPath);
    listInsert(useDllDirectories(), string(dirPath));

    resolveDependencies(list, pePath);

    int longestName = 4;
    int longestPath = 4;

    for (Deferable **item = list->data; *item != NULL; item++) {
        DllDependency *dependency = *(DllDependency **) item;
        longestName = max(longestName, dependency->dllName->length);
        longestPath = max(longestPath, dependency->dllPath->length);
    }

    VHContext ctx = {
            .file_descriptor = 1,
            .indent = arguments->indent,
            .color_output = arguments->colorOutput
    };

    if (arguments->json) vhi_object_start(&ctx);
    else if (!arguments->listPaths && !arguments->listNames) {
        if (arguments->colorOutput) printf("\033[1m");

        printf("%-*s   %-*s   %-7s   %s\n",
               longestName, "Name", longestPath, "Path",
               "System?", "Delayed?");

        if (arguments->colorOutput) printf("\033[0m");
    }

    for (Deferable **item = list->data; *item != NULL; item++) {
        DllDependency *dependency = *(DllDependency **) item;

        if (arguments->noSystem && dependency->isSystem) continue;
        if (arguments->onlySystem && !dependency->isSystem) continue;

        if (arguments->json) {
            vhi_key_sized(&ctx, SACL(dependency->dllName));

            vhi_object(&ctx, {
                vhi_key(&ctx, "name");
                vhi_string_sized(&ctx, SACL(dependency->dllName));
                vhi_key(&ctx, "path");
                if (dependency->dllPath->cStr == UnknownLoc) vhi_null(&ctx);
                else vhi_string_sized(&ctx, SACL(dependency->dllPath));
                vhi_key(&ctx, "is_system");
                vhi_bool(&ctx, dependency->isSystem);
                vhi_key(&ctx, "is_delayed");
                vhi_bool(&ctx, dependency->isDelayed);
                vhi_key(&ctx, "symbols");

                vhi_array(&ctx, {
                        for (int index = 0; index < dependency->symbols->length; index++) {
                            vhi_string_sized(&ctx, SACL(dependency->symbols->data[index]));
                        }
                });
            });
        } else if (arguments->listPaths) {
            printf(SFMT "\n", SALC(dependency->dllPath));
        } else if (arguments->listNames) {
            printf(SFMT "\n", SALC(dependency->dllName));
        } else if (arguments->colorOutput) {
            printf("\033[36m%-*.*s\033[0m   %-*.*s   %18s   %s\n",
                   longestName, SALC(dependency->dllName),
                   longestPath, SALC(dependency->dllPath),
                   dependency->isSystem ? "\033[1;92mtrue\033[0m" : "\033[1;91mfalse\033[0m",
                   dependency->isDelayed ? "\033[1;92mtrue\033[0m" : "\033[1;91mfalse\033[0m");
        } else {
            printf("%-*.*s   %-*.*s   %7s   %s\n",
                   longestName, SALC(dependency->dllName),
                   longestPath, SALC(dependency->dllPath),
                   dependency->isSystem ? "true" : "false",
                   dependency->isDelayed ? "true" : "false");
        }
    }

    if (arguments->json) {
        vhi_object_end(&ctx);
        printf("\n");
    }

    return 0;
}
