#include <string.h>
#include <slcurses.h>
#include "ordered_unique_list.h"

static void checkCapacity(OrderedUniqueList *list) {
    if (list->length == list->capacity) {
        list->capacity *= 2;
        list->data = reallocarray(list->data, list->capacity, sizeof(void *));

        for (int i = list->length; i < list->capacity; i++) {
            list->data[i] = NULL;
        }
    }
}

OrderedUniqueList *newOrderedUniqueList(ComparatorFunc comparator) {
    OrderedUniqueList *list = dalloc(OrderedUniqueList);
    list->comparator = comparator;
    list->capacity = 16;
    list->data = calloc(list->capacity, sizeof(void *));
    return list;
}

bool listInsertOrFetch(OrderedUniqueList *list, void *dataPtr) {
    void **data = dataPtr;
    checkCapacity(list);

    int index = 0;
    if (listSearch(list, *data, &index)) {
        deferFunc(data);
        *data = list->data[index];
        return false;
    }

    for (int next = list->length; next > index; next--) {
        list->data[next] = list->data[next - 1];
    }

    list->data[index] = *data;
    list->length++;
    return true;
}

bool listInsert(OrderedUniqueList *list, void *data) {
    checkCapacity(list);

    int index = 0;
    if (listSearch(list, data, &index)) return false;

    for (int next = list->length; next > index; next--) {
        list->data[next] = list->data[next - 1];
    }

    list->data[index] = data;
    list->length++;
    return true;
}

static bool listSearchImpl(OrderedUniqueList *list, Deferable *data, int *index, int high, int low) {
    if (high < low || *index < 0 || *index >= list->length) {
        *index = clamp(0, list->length, *index);
        return false;
    }

    int comparison = list->comparator(list->data[*index], data);

    if (comparison > 0) high = *index - 1;
    else if (comparison < 0) low = *index + 1;
    else return true;

    *index = low + ((high - low) / 2);
    return listSearchImpl(list, data, index, high, low);
}

bool listSearch(OrderedUniqueList *list, Deferable *data, int *index) {
    if (index == NULL) {
        int dIndex = list->length / 2;
        return listSearchImpl(list, data, &dIndex, list->length, -1);
    }

    *index = list->length / 2;
    return listSearchImpl(list, data, index, list->length, -1);
}

Deferable *listFind(OrderedUniqueList *list, Deferable *data) {
    int index = 0;
    if (!listSearch(list, data, &index)) return NULL;
    return list->data[index];
}

void copyOrderedUniqueList(OrderedUniqueList *oldList, OrderedUniqueList *newList) {
    newList->capacity = oldList->capacity;
    newList->length = oldList->length;
    newList->comparator = oldList->comparator;
    newList->data = calloc(newList->length, sizeof(Deferable *));

    for (int index = 0; index < oldList->length; index++) {
        newList->data[index] = oldList->data[index];
    }
}

void freeOrderedUniqueList(OrderedUniqueList *list) {
    for (int index = 0; index < list->length; index++) {
        deferFunc(&list->data[index]);
    }

    if (list->data != NULL) free(list->data);
}
