#include <stddef.h>
#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <stdbool.h>
#include <stdlib.h>
#include <ctype.h>

#ifndef VH_MAX_DEPTH
#define VH_MAX_DEPTH 32
#endif

#ifndef VH_DEFAULT_LIST_CAP
#define VH_DEFAULT_LIST_CAP 16
#endif

#ifndef VH_DEF
#define VH_DEF
#endif

#ifndef VH_HEADER
#define VH_HEADER

typedef void(*VHWriter)(void *user_data, const char *data, size_t length);

typedef struct {
    size_t start_index, end_index;
    const char *data;
} VHStringBuffer;

typedef enum {
    VHNull,
    VHBoolean,
    VHInteger,
    VHDouble,
    VHString,
    VHArray,
    VHObject,
    VHObjectKey
} VHType;

typedef struct VHElement {
    VHType type;
    size_t size;
    size_t index;

    union {
        bool boolean;
        int64_t integer;
        double number;
        const char *string;
        struct VHElement *elements;
    };
} VHElement;

typedef struct {
    int file_descriptor, indent, level;
    bool color_output;
    VHType level_types[VH_MAX_DEPTH];
    VHType last_type;
    VHWriter writer;
    void *user_data;
} VHContext;

#define vhi_object(context, body) do { \
    vhi_object_start((context));       \
    body                              \
    vhi_object_end((context));         \
} while(0)

#define vhi_array(context, body) do { \
    vhi_array_start((context));       \
    body                             \
    vhi_array_end((context));         \
} while(0)

#define vhr_true ((VHElement){.type=VHBoolean, .boolean=true})
#define vhr_false ((VHElement){.type=VHBoolean, .boolean=false})
#define vhr_null ((VHElement){.type=VHNull})

VH_DEF const char *vh_type_str(VHType type);

VH_DEF VHStringBuffer vhsb(const char *string);

VH_DEF VHStringBuffer vhsb_sized(const char *string, size_t string_length);

VH_DEF const char *vhsb_start(VHStringBuffer *buffer);

VH_DEF size_t vhsb_length(VHStringBuffer *buffer);

VH_DEF void vhsb_trim(VHStringBuffer *buffer);

VH_DEF void vhsb_trim_left(VHStringBuffer *buffer);

VH_DEF void vhsb_trim_right(VHStringBuffer *buffer);

VH_DEF bool vhsb_peek(VHStringBuffer *buffer, const char *string);

VH_DEF bool vhsb_peek_sized(VHStringBuffer *buffer, const char *string, size_t string_length);

VH_DEF bool vhsb_expect(VHStringBuffer *buffer, const char *string);

VH_DEF bool vhsb_expect_sized(VHStringBuffer *buffer, const char *string, size_t string_length);

VH_DEF VHStringBuffer vhsb_chop_delimiter(VHStringBuffer *buffer, char delimiter);

VH_DEF void vhi_write(const VHContext *context, const char *data, size_t length);

VH_DEF void vhi_write_escaped(const VHContext *context, const char *data, size_t length);

VH_DEF void vhi_write_integer(const VHContext *context, int64_t integer);

VH_DEF void vhi_write_separation(const VHContext *context);

VH_DEF void vhi_write_indentation(const VHContext *context);

VH_DEF void vhi_object_start(VHContext *context);

VH_DEF void vhi_object_end(VHContext *context);

VH_DEF void vhi_array_start(VHContext *context);

VH_DEF void vhi_array_end(VHContext *context);

VH_DEF void vhi_key(VHContext *context, const char *key);

VH_DEF void vhi_key_sized(VHContext *context, const char *key, size_t key_length);

VH_DEF void vhi_string(VHContext *context, const char *string);

VH_DEF void vhi_string_sized(VHContext *context, const char *string, size_t string_length);

VH_DEF void vhi_integer(VHContext *context, int64_t integer);

VH_DEF void vhi_double(VHContext *context, double number);

VH_DEF void vhi_null(VHContext *context);

VH_DEF void vhi_bool(VHContext *context, bool boolean);

VH_DEF VHElement vhr_object();

VH_DEF VHElement vhr_array();

VH_DEF VHElement vhr_string(const char *string);

VH_DEF VHElement vhr_string_sized(const char *string, size_t string_length);

VH_DEF VHElement vhr_integer(int64_t integer);

VH_DEF VHElement vhr_double(double number);

VH_DEF VHElement vhr_boolean(bool boolean);

VH_DEF void vhr_object_append(VHElement *object, const char *key, VHElement value);

VH_DEF void vhr_object_append_sized(VHElement *object, const char *key, size_t key_length, VHElement value);

VH_DEF void vhr_array_append(VHElement *array, VHElement element);

VH_DEF void vhr_format(VHContext *context, VHElement *data);

VH_DEF void vhr_free(VHElement *data);

VH_DEF VHType vhi_peek_type(VHStringBuffer *buffer);

VH_DEF bool vhi_expect_object_start(VHStringBuffer *buffer);

VH_DEF bool vhi_expect_object_end(VHStringBuffer *buffer);

VH_DEF bool vhi_expect_array_start(VHStringBuffer *buffer);

VH_DEF bool vhi_expect_array_end(VHStringBuffer *buffer);

VH_DEF bool vhi_expect_string(VHStringBuffer *buffer, const char **key, size_t *key_length);

VH_DEF bool vhi_expect_key(VHStringBuffer *buffer, const char **key, size_t *key_length);

VH_DEF bool vhi_expect_integer(VHStringBuffer *buffer, int64_t *integer);

VH_DEF bool vhi_expect_double(VHStringBuffer *buffer, double *number);

VH_DEF bool vhi_expect_boolean(VHStringBuffer *buffer, bool *boolean);

VH_DEF bool vhi_expect_null(VHStringBuffer *buffer);

VH_DEF bool vhr_parse(VHStringBuffer *buffer, VHElement *element);

#endif

#ifdef VH_IMPLEMENTATION

VH_DEF const char *vh_type_str(VHType type) {
    switch (type) {
        case VHNull:
            return "VHNull";
        case VHBoolean:
            return "VHBoolean";
        case VHInteger:
            return "VHInteger";
        case VHDouble:
            return "VHDouble";
        case VHString:
            return "VHString";
        case VHArray:
            return "VHArray";
        case VHObject:
            return "VHObject";
        case VHObjectKey:
            return "VHObjectKey";
        default:
            assert(0);
    }
}

VH_DEF VHStringBuffer vhsb(const char *string) {
    return vhsb_sized(string, strlen(string));
}

VH_DEF VHStringBuffer vhsb_sized(const char *string, size_t string_length) {
    return (VHStringBuffer) {
            .start_index = 0,
            .end_index = string_length - 1,
            .data = string
    };
}

VH_DEF const char *vhsb_start(VHStringBuffer *buffer) {
    return &buffer->data[buffer->start_index];
}

VH_DEF size_t vhsb_length(VHStringBuffer *buffer) {
    return buffer->end_index - buffer->start_index + 1;
}

VH_DEF void vhsb_trim(VHStringBuffer *buffer) {
    vhsb_trim_left(buffer);
    vhsb_trim_right(buffer);
}

VH_DEF void vhsb_trim_left(VHStringBuffer *buffer) {
    for (; isspace(buffer->data[buffer->start_index]) &&
           buffer->start_index <= buffer->end_index; buffer->start_index++);
}

VH_DEF void vhsb_trim_right(VHStringBuffer *buffer) {
    for (; isspace(buffer->data[buffer->end_index]) && buffer->start_index <= buffer->end_index; buffer->end_index--);
}


VH_DEF bool vhsb_peek(VHStringBuffer *buffer, const char *string) {
    return vhsb_peek_sized(buffer, string, strlen(string));
}

VH_DEF bool vhsb_peek_sized(VHStringBuffer *buffer, const char *string, size_t string_length) {
    return memcmp(vhsb_start(buffer), string, string_length) == 0;
}

VH_DEF bool vhsb_expect(VHStringBuffer *buffer, const char *string) {
    return vhsb_expect_sized(buffer, string, strlen(string));
}

VH_DEF bool vhsb_expect_sized(VHStringBuffer *buffer, const char *string, size_t string_length) {
    if (vhsb_peek_sized(buffer, string, string_length)) {
        buffer->start_index += string_length;
        return true;
    }

    return false;
}

VH_DEF VHStringBuffer vhsb_chop_delimiter(VHStringBuffer *buffer, char delimiter) {
    VHStringBuffer copy = {
            .start_index = buffer->start_index,
            .data = buffer->data
    };

    for (; buffer->data[buffer->start_index] != delimiter &&
           buffer->start_index <= buffer->end_index; buffer->start_index++);

    copy.end_index = buffer->start_index - 1;
    buffer->start_index++;
    return copy;
}

VH_DEF void vhi_write(const VHContext *context, const char *data, size_t length) {
    if (context->file_descriptor) {
        write(context->file_descriptor, data, length);
    } else if (context->writer != NULL) {
        context->writer(context->user_data, data, length);
    }
}

VH_DEF void vhi_write_escaped(const VHContext *context, const char *data, size_t length) {
    char escaped_data[length * 2];
    size_t escaped_data_length = 0;

    for (size_t index = 0; index < length; index++) {
        char current = data[index];

        if (current == '"' || current == '\\' || current == '\b' || current == '\f' || current == '\n' ||
            current == '\r' || current == '\t') {
            escaped_data[escaped_data_length++] = '\\';
            escaped_data[escaped_data_length++] = current;
        } else if (current < 32) {
            assert(0 && "Not implemented yet");
        } else {
            escaped_data[escaped_data_length++] = current;
        }
    }

    vhi_write(context, escaped_data, escaped_data_length);
}

VH_DEF void vhi_write_integer(const VHContext *context, int64_t integer) {
    if (context->color_output) vhi_write(context, "\033[1;34m", 7);

    if (integer < 0) {
        vhi_write(context, "-", 1);
        integer = -integer;
    }

    int64_t largest_divisor = 1;
    while (integer / largest_divisor > 9) largest_divisor *= 10;

    for (; largest_divisor > 0; largest_divisor /= 10) {
        int64_t digit = integer / largest_divisor;
        integer -= digit * largest_divisor;
        char digit_chr = (char) (digit + '0');
        vhi_write(context, &digit_chr, 1);
    }

    if (context->color_output) vhi_write(context, "\033[0m", 4);
}

VH_DEF void vhi_write_separation(const VHContext *context) {
    if ((context->level_types[context->level] == VHObject && context->last_type != VHObject &&
         context->last_type != VHObjectKey) ||
        (context->level_types[context->level] == VHArray && context->last_type != VHArray)) {

        if (context->color_output) vhi_write(context, "\033[1;37m,\033[0m", 12);
        else vhi_write(context, ",", 1);
    }

    if (context->level != 0) {
        vhi_write_indentation(context);
    }
}

VH_DEF void vhi_write_indentation(const VHContext *context) {
    if (context->indent <= 0) {
        return;
    }

    if (context->last_type == VHObjectKey) {
        vhi_write(context, " ", 1);
    } else {
        vhi_write(context, "\n", 1);

        size_t indentation_length = context->level * context->indent;
        char indentation[indentation_length];
        memset(indentation, ' ', indentation_length);
        vhi_write(context, indentation, indentation_length);
    }
}

VH_DEF void vhi_object_start(VHContext *context) {
    assert(context->level <= VH_MAX_DEPTH);
    vhi_write_separation(context);

    if (context->color_output) vhi_write(context, "\033[1;37m{\033[0m", 12);
    else vhi_write(context, "{", 1);

    context->level_types[++context->level] = VHObject;
    context->last_type = VHObject;
}

VH_DEF void vhi_object_end(VHContext *context) {
    context->level--;
    vhi_write_indentation(context);

    if (context->color_output) vhi_write(context, "\033[1;37m}\033[0m", 12);
    else vhi_write(context, "}", 1);
}

VH_DEF void vhi_array_start(VHContext *context) {
    assert(context->level <= VH_MAX_DEPTH);
    vhi_write_separation(context);

    if (context->color_output) vhi_write(context, "\033[1;37m[\033[0m", 12);
    else vhi_write(context, "[", 1);

    context->level_types[++context->level] = VHArray;
    context->last_type = VHArray;
}

VH_DEF void vhi_array_end(VHContext *context) {
    context->level--;
    vhi_write_indentation(context);

    if (context->color_output) vhi_write(context, "\033[1;37m]\033[0m", 12);
    else vhi_write(context, "]", 1);
}

VH_DEF void vhi_key(VHContext *context, const char *key) {
    vhi_key_sized(context, key, strlen(key));
}

VH_DEF void vhi_key_sized(VHContext *context, const char *key, size_t key_length) {
    vhi_write_separation(context);

    if (context->color_output) vhi_write(context, "\033[1;32m\"", 8);
    else vhi_write(context, "\"", 1);

    vhi_write_escaped(context, key, key_length);

    if (context->color_output) vhi_write(context, "\"\033[1;39m:\033[0m", 13);
    else vhi_write(context, "\":", 2);

    context->last_type = VHObjectKey;
}

VH_DEF void vhi_string(VHContext *context, const char *string) {
    vhi_string_sized(context, string, strlen(string));
}

VH_DEF void vhi_string_sized(VHContext *context, const char *string, size_t string_length) {
    vhi_write_separation(context);

    if (context->color_output) vhi_write(context, "\033[32m\"", 6);
    else vhi_write(context, "\"", 1);

    vhi_write_escaped(context, string, string_length);

    if (context->color_output) vhi_write(context, "\"\033[0m", 5);
    else vhi_write(context, "\"", 1);

    context->last_type = VHString;
}

VH_DEF void vhi_integer(VHContext *context, int64_t integer) {
    vhi_write_separation(context);
    vhi_write_integer(context, integer);
    context->last_type = VHInteger;
}

VH_DEF void vhi_double(VHContext *context, double number) {
    vhi_write_separation(context);

    if (context->color_output) vhi_write(context, "\033[1;34m", 7);

    int64_t decimal = (int64_t) number;
    int64_t fraction = 0, fraction_counter = 0;
    number -= (double) decimal;

    if (number < 0.0) {
        number = -number;
    }

    do {
        fraction = fraction * 10 + (int64_t) number;
        number = (number - (double) (int64_t) number) * 10;
    } while (number != 0.0 && fraction_counter++ < 10);

    vhi_write_integer(context, decimal);
    vhi_write(context, ".", 1);
    vhi_write_integer(context, fraction);
    context->last_type = VHDouble;

    if (context->color_output) vhi_write(context, "\033[0m", 4);
}

VH_DEF void vhi_null(VHContext *context) {
    vhi_write_separation(context);

    if (context->color_output) vhi_write(context, "\033[1;33mnull\033[1m", 15);
    else vhi_write(context, "null", 4);

    context->last_type = VHNull;
}

VH_DEF void vhi_bool(VHContext *context, bool boolean) {
    vhi_write_separation(context);

    if (boolean) {
        if (context->color_output) vhi_write(context, "\033[1;92mtrue\033[0m", 15);
        else vhi_write(context, "true", 4);
    } else {
        if (context->color_output) vhi_write(context, "\033[1;91mfalse\033[0m", 16);
        else vhi_write(context, "false", 5);
    }

    context->last_type = VHBoolean;
}

VH_DEF VHElement vhr_object() {
    return (VHElement) {
            .type = VHObject,
            .size = VH_DEFAULT_LIST_CAP,
            .index = 0,
            .elements = calloc(VH_DEFAULT_LIST_CAP, 2 * sizeof(VHElement))
    };
}

VH_DEF VHElement vhr_array() {
    return (VHElement) {
            .type = VHArray,
            .size = VH_DEFAULT_LIST_CAP,
            .elements = calloc(VH_DEFAULT_LIST_CAP, sizeof(VHElement))
    };
}

VH_DEF VHElement vhr_string(const char *string) {
    return vhr_string_sized(string, strlen(string));
}

VH_DEF VHElement vhr_string_sized(const char *string, size_t string_length) {
    return (VHElement) {
            .type = VHString,
            .size = string_length,
            .string = string
    };
}

VH_DEF VHElement vhr_integer(int64_t integer) {
    return (VHElement) {
            .type = VHInteger,
            .integer = integer
    };
}

VH_DEF VHElement vhr_double(double number) {
    return (VHElement) {
            .type = VHDouble,
            .number = number
    };
}

VH_DEF VHElement vhr_boolean(bool boolean) {
    return (VHElement) {
            .type = VHBoolean,
            .boolean = boolean
    };
}

VH_DEF void vhr_object_append(VHElement *object, const char *key, VHElement value) {
    vhr_object_append_sized(object, key, strlen(key), value);
}

VH_DEF void vhr_object_append_sized(VHElement *object, const char *key, size_t key_length, VHElement value) {
    assert(object->type == VHObject);

    if (object->index >= object->size) {
        object->size *= 2;
        object->elements = reallocarray(object->elements, object->size, sizeof(VHElement));
    }

    object->elements[object->index++] = (VHElement) {
            .type = VHObjectKey,
            .string = key,
            .size = key_length
    };

    object->elements[object->index++] = value;
}

VH_DEF void vhr_array_append(VHElement *array, VHElement element) {
    assert(array->type == VHArray);

    if (array->index > array->size) {
        array->size *= 2;
        array->elements = reallocarray(array->elements, array->size, sizeof(VHElement));
    }

    array->elements[array->index++] = element;
}

VH_DEF void vhr_format(VHContext *context, VHElement *data) {
    switch (data->type) {
        case VHNull:
            vhi_null(context);
            break;
        case VHBoolean:
            vhi_bool(context, data->boolean);
            break;
        case VHInteger:
            vhi_integer(context, data->integer);
            break;
        case VHDouble:
            vhi_double(context, data->number);
            break;
        case VHString:
            vhi_string_sized(context, data->string, data->size);
            break;
        case VHArray: {
            vhi_array_start(context);

            for (size_t i = 0; i < data->index; i++) {
                vhr_format(context, &data->elements[i]);
            }

            vhi_array_end(context);
        }
            break;
        case VHObject: {
            vhi_object_start(context);

            for (size_t i = 0; i < data->index; i++) {
                vhr_format(context, &data->elements[i]);
            }

            vhi_object_end(context);
        }
            break;
        case VHObjectKey:
            vhi_key_sized(context, data->string, data->size);
            break;
        default:
            assert(0 && "Unreachable");
    }

    if (context->level == 0 && context->indent > 0) {
        vhi_write(context, "\n", 1);
    }
}

VH_DEF void vhr_free(VHElement *data) {
    if (data->type == VHArray || data->type == VHObject) {
        for (size_t index = 0; index < data->index; index++) {
            vhr_free(&data->elements[index]);
        }

        free(data->elements);
    }
}

VH_DEF VHType vhi_peek_type(VHStringBuffer *buffer) {
    switch (buffer->data[buffer->start_index]) {
        case '{':
        case '}':
            return VHObject;
        case '[':
        case ']':
            return VHArray;
        case '"':
            return VHString;
        case 't':
        case 'f':
            return VHBoolean;
        case 'n':
            return VHNull;
        case '-':
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            for (size_t i = buffer->start_index + 1; i <= buffer->end_index; i++) {
                if (buffer->data[i] == '.') {
                    return VHDouble;
                } else if (buffer->data[i] < '0' || buffer->data[i] > '9') {
                    return VHInteger;
                }
            }

            return VHInteger;
        default:
            assert(0 && buffer->data[buffer->start_index]);
    }
}

VH_DEF bool vhi_expect_object_start(VHStringBuffer *buffer) {
    if (!vhsb_expect_sized(buffer, "{", 1)) {
        return false;
    }

    vhsb_trim_left(buffer);
    return true;
}

VH_DEF bool vhi_expect_object_end(VHStringBuffer *buffer) {
    if (!vhsb_expect_sized(buffer, "}", 1)) {
        return false;
    }

    vhsb_trim_left(buffer);

    if (vhsb_expect_sized(buffer, ",", 1)) {
        vhsb_trim_left(buffer);
    }

    return true;
}

VH_DEF bool vhi_expect_array_start(VHStringBuffer *buffer) {
    if (!vhsb_expect_sized(buffer, "[", 1)) {
        return false;
    }

    vhsb_trim_left(buffer);
    return true;
}

VH_DEF bool vhi_expect_array_end(VHStringBuffer *buffer) {
    if (!vhsb_expect_sized(buffer, "]", 1)) {
        return false;
    }

    vhsb_trim_left(buffer);

    if (vhsb_expect_sized(buffer, ",", 1)) {
        vhsb_trim_left(buffer);
    }

    return true;
}

VH_DEF bool vhi_expect_string(VHStringBuffer *buffer, const char **key, size_t *key_length) {
    if (!vhsb_expect_sized(buffer, "\"", 1)) {
        return false;
    }

    VHStringBuffer keyBuffer = vhsb_chop_delimiter(buffer, '"');
    vhsb_trim_left(buffer);

    if (vhsb_expect_sized(buffer, ",", 1)) {
        vhsb_trim_left(buffer);
    }

    *key = vhsb_start(&keyBuffer);
    *key_length = vhsb_length(&keyBuffer);
    return true;
}

VH_DEF bool vhi_expect_key(VHStringBuffer *buffer, const char **key, size_t *key_length) {
    if (!vhsb_expect_sized(buffer, "\"", 1)) {
        return false;
    }

    VHStringBuffer key_buffer = vhsb_chop_delimiter(buffer, '"');
    vhsb_trim_left(buffer);

    if (!vhsb_expect_sized(buffer, ":", 1)) {
        return false;
    }

    vhsb_trim_left(buffer);

    *key = vhsb_start(&key_buffer);
    *key_length = vhsb_length(&key_buffer);
    return true;
}

VH_DEF bool vhi_expect_integer(VHStringBuffer *buffer, int64_t *integer) {
    *integer = 0;
    VHStringBuffer integer_buffer = vhsb_chop_delimiter(buffer, ',');
    vhsb_trim_left(buffer);

    bool negative = vhsb_expect_sized(&integer_buffer, "-", 1);

    for (; integer_buffer.start_index <= integer_buffer.end_index; integer_buffer.start_index++) {
        char current = integer_buffer.data[integer_buffer.start_index];

        if (current < '0' || current > '9') {
            break;
        }

        *integer = *integer * 10 + (int64_t) (current - '0');
    }

    if (vhsb_expect_sized(&integer_buffer, "e", 1) || vhsb_expect_sized(&integer_buffer, "E", 1)) {
        bool negative_exponent = false;
        int64_t exponent = 0;

        if (vhsb_expect_sized(&integer_buffer, "-", 1)) {
            negative_exponent = true;
        } else vhsb_expect_sized(&integer_buffer, "+", 1);

        for (; integer_buffer.start_index <= integer_buffer.end_index; integer_buffer.start_index++) {
            char current = integer_buffer.data[integer_buffer.start_index];

            if (current < '0' || current > '9') {
                return false;
            }

            exponent = exponent * 10 + (int64_t) (current - '0');
        }

        for (int64_t i = 0; i < exponent; i++) {
            if (negative_exponent) {
                *integer /= 10;
            } else {
                *integer *= 10;
            }
        }
    }

    if (negative) {
        *integer = -*integer;
    }

    return true;
}

VH_DEF bool vhi_expect_double(VHStringBuffer *buffer, double *number) {
    VHStringBuffer integer_buffer = vhsb_chop_delimiter(buffer, '.');

    bool negative = vhsb_expect_sized(buffer, "-", 1);
    double decimal = 0, fraction = 0;

    for (; integer_buffer.start_index <= integer_buffer.end_index; integer_buffer.start_index++) {
        char current = integer_buffer.data[integer_buffer.start_index];

        if (current < '0' || current > '9') {
            break;
        }

        decimal = decimal * 10 + (double) (current - '0');
    }

    VHStringBuffer fraction_buffer = vhsb_chop_delimiter(buffer, ',');
    vhsb_trim_left(buffer);

    for (; fraction_buffer.end_index >= fraction_buffer.start_index; fraction_buffer.end_index--) {
        char current = fraction_buffer.data[fraction_buffer.end_index];

        if (current < '0' || current > '9') {
            break;
        }

        fraction = (fraction + (double) (current - '0')) / 10;
    }

    *number = decimal + fraction;

    if (vhsb_expect_sized(&integer_buffer, "e", 1) || vhsb_expect_sized(&integer_buffer, "E", 1)) {
        bool negative_exponent = false;
        int64_t exponent = 0;

        if (vhsb_expect_sized(&integer_buffer, "-", 1)) {
            negative_exponent = true;
        } else vhsb_expect_sized(&integer_buffer, "+", 1);

        for (; integer_buffer.start_index <= integer_buffer.end_index; integer_buffer.start_index++) {
            char current = integer_buffer.data[integer_buffer.start_index];

            if (current < '0' || current > '9') {
                return false;
            }

            exponent = exponent * 10 + (int64_t) (current - '0');
        }

        for (int64_t i = 0; i < exponent; i++) {
            if (negative_exponent) {
                *number /= 10;
            } else {
                *number *= 10;
            }
        }
    }

    if (negative) {
        *number = -*number;
    }

    return true;
}

VH_DEF bool vhi_expect_boolean(VHStringBuffer *buffer, bool *boolean) {
    if (vhsb_expect_sized(buffer, "true", 4)) {
        *boolean = true;
    } else if (vhsb_expect_sized(buffer, "false", 5)) {
        *boolean = false;
    } else {
        return false;
    }

    vhsb_trim_left(buffer);

    if (vhsb_expect_sized(buffer, ",", 1)) {
        vhsb_trim_left(buffer);
    }

    return true;
}

VH_DEF bool vhi_expect_null(VHStringBuffer *buffer) {
    if (!vhsb_expect_sized(buffer, "null", 4)) {
        return false;
    }

    vhsb_trim_left(buffer);

    if (vhsb_expect_sized(buffer, ",", 1)) {
        vhsb_trim_left(buffer);
    }

    return true;
}

VH_DEF bool vhr_parse(VHStringBuffer *buffer, VHElement *element) {
    switch (vhi_peek_type(buffer)) {
        case VHObject: {
            if (!vhi_expect_object_start(buffer)) {
                return false;
            }

            *element = vhr_object();

            while (vhi_peek_type(buffer) != VHObject) {
                const char *key;
                size_t key_length = {0};
                VHElement value = {0};

                if (!vhi_expect_key(buffer, &key, &key_length)) {
                    return false;
                }

                if (!vhr_parse(buffer, &value)) {
                    if (value.type == VHObject || value.type == VHArray) {
                        vhr_free(&value);
                    }

                    return false;
                }

                vhr_object_append_sized(element, key, key_length, value);
            }

            if (!vhi_expect_object_end(buffer)) {
                return false;
            }
        }
            break;
        case VHArray: {
            if (!vhi_expect_array_start(buffer)) {
                return false;
            }

            *element = vhr_array();

            while (vhi_peek_type(buffer) != VHArray) {
                VHElement value;

                if (!vhr_parse(buffer, &value)) {
                    if (value.type == VHObject || value.type == VHArray) {
                        vhr_free(&value);
                    }

                    return false;
                }

                vhr_array_append(element, value);
            }

            if (!vhi_expect_array_end(buffer)) {
                return false;
            }
        }
            break;
        case VHString: {
            const char *string;
            size_t string_length;

            if (!vhi_expect_string(buffer, &string, &string_length)) {
                return false;
            }

            *element = vhr_string_sized(string, string_length);
        }
            break;
        case VHInteger: {
            int64_t integer;

            if (!vhi_expect_integer(buffer, &integer)) {
                return false;
            }

            *element = vhr_integer(integer);
        }
            break;
        case VHDouble: {
            double number;

            if (!vhi_expect_double(buffer, &number)) {
                return false;
            }

            *element = vhr_double(number);
        }
            break;
        case VHBoolean: {
            bool boolean;

            if (!vhi_expect_boolean(buffer, &boolean)) {
                return false;
            }

            *element = vhr_boolean(boolean);
        }
            break;
        case VHNull: {
            if (!vhi_expect_null(buffer)) {
                return false;
            }

            *element = vhr_null;
        }
            break;
        default:
            return false;
    }

    return true;
}

#endif
